/*
 *   timer.c
 *
 *   This file is part of Emu48
 *
 *   Copyright (C) 1995 Sebastien Carlier
 *
 */
#include "pch.h"
#include "Emu48.h"

#define AUTO_OFF    10						// 25.02.98 cg, new, Time in minutes for 'auto off'

// 25.02.98 cg, new, Ticks for 01.01.1970 00:00:00
#define UNIX_0_TIME	0x0001cf2e8f800000

// 25.02.98 cg, new, Ticks for 'auto off'
#define OFF_TIME	((LONGLONG) (AUTO_OFF * 60) << 13)

// 06.10.97 cg, new, memory adresses for clock
#define ACCESSTIME	((cCurrentRomType=='S')?0x70052:0x80058)
#define ACCESSCRC	((cCurrentRomType=='S')?0x7005F:0x80065)
#define TIMEOUT		((cCurrentRomType=='S')?0x70063:0x80069)
#define TIMEOUTCLK	((cCurrentRomType=='S')?0x70070:0x80076)

#define T2_FREQ		8192					// 06.04.98 cg, new, Timer2 frequency

// 27.02.98 cg, new, Timer definitions
#define IO_OFFSET	0x100					// Mapping offset

// Timer addresses
#define TIMER1_CTRL	(0x12e - IO_OFFSET)		// Timer1 Control
#define TIMER2_CTRL	(0x12f - IO_OFFSET)		// Timer2 Control
#define TIMER1		(0x137 - IO_OFFSET)		// Timer1 (4 bit)
#define TIMER2		(0x13f - IO_OFFSET)		// Timer2 (32 bit, LSB first)

// 0x12e Timer1 Control [SRQ WKE INT XTRA]
#define SRQ			0x08					// Service request
#define WKE			0x04					// Wake up
#define INT			0x02					// Interrupt
#define XTRA		0x01					// Extra function

// 0x12f Timer1 Control [SRQ WKE INT RUN]
#define SRQ			0x08					// Service request
#define WKE			0x04					// Wake up
#define INT			0x02					// Interrupt
#define RUN			0x01					// Timer run

BOOL bAccurateTimer = TRUE;
UINT uT1Period  = 62;

static BOOL  bStarted   = FALSE;
static BOOL  bOutRange  = FALSE;			// 21.04.98 cg, new, flag if timer value out of range
static UINT  uT1TimerId = 0;
static UINT  uT2TimerId = 0;
// static DWORD dwT1Ticks = 0;				// 19.02.98 cg, removed, not used
// static DWORD dwT2Init  = 0;				// 06.04.98 cg, removed, not used any more
// static DWORD dwT2Step  = 0;				// 06.04.98 cg, removed, not used any more
// static DWORD dwT2Ticks = 0;				// 06.04.98 cg, removed, not used any more

static LARGE_INTEGER lFreq;					// 06.04.98 cg, new, counter frequency
static LARGE_INTEGER lT2Ref;				// 06.04.98 cg, new, counter value at timer2 start

static TIMECAPS tc;							// 21.04.98 cg, new, timer information

static CRITICAL_SECTION csT1Lock;			// 21.04.98 cg, new, critical section for timer1 access
static CRITICAL_SECTION csT2Lock;			// 21.04.98 cg, new, critical section for timer2 access

static __inline MAX(int a, int b) {return (a>b)?a:b;} 

static void CALLBACK TimeProc(UINT uEventId, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2);

static VOID SetAccesstime(VOID)				// 06.10.97 cg, new, set date and time
{
	SYSTEMTIME ts;
	LONGLONG   ticks, time;
	DWORD      dw;
	WORD       crc, i;
	BYTE       p[4];

	_ASSERT(sizeof(LONGLONG) == 8);			// check size of datatype

	GetLocalTime(&ts);						// local time, _ftime() cause memory/resource leaks

	// calculate days until 01.01.1970
	dw = (DWORD) ts.wMonth;
	if (dw > 2)
		dw -= 3L;
	else
	{
		dw += 9L;
		--ts.wYear;
	}
	dw = (DWORD) ts.wDay + (153L * dw + 2L) / 5L;
	dw += (146097L * (((DWORD) ts.wYear) / 100L)) / 4L;
	dw +=   (1461L * (((DWORD) ts.wYear) % 100L)) / 4L;
	dw -= 719469L;

	// convert into seconds and add time
	dw = dw * 24L + (DWORD) ts.wHour;
	dw = dw * 60L + (DWORD) ts.wMinute;
	dw = dw * 60L + (DWORD) ts.wSecond;

	// create timerticks = (s + ms) * 8192
	ticks = ((LONGLONG) dw << 13) | (((LONGLONG) ts.wMilliseconds << 10) / 125);

	ticks += UNIX_0_TIME;					// add offset ticks from year 0
	ticks += (LONG) Chipset.t2;				// add actual timer2 value

	time = ticks;							// save for calc. timeout
	time += OFF_TIME;						// add 10 min for auto off

	crc = 0x0;								// reset crc value
	dw = ACCESSTIME;						// HP adresses for clock

	for (i = 0; i < 13; ++i, ++dw)			// write date and time
	{
		*p = (BYTE) ticks & 0xf;
		crc = (crc >> 4) ^ (((crc ^ ((WORD) *p)) & 0xf) * 0x1081);
		Nwrite(p,dw,1);
		ticks >>= 4;
	}

	Nunpack(p,crc,4);						// write crc
	Nwrite(p,ACCESSCRC,4);

	dw = TIMEOUT;							// HP adresses for timeout

	for (i = 0; i < 13; ++i, ++dw)			// write time for auto off
	{
		*p = (BYTE) time & 0xf;
		Nwrite(p,dw,1);
		time >>= 4;
	}

	*p = 0xf;
	Nwrite(p,TIMEOUTCLK,1);
}

static DWORD CalcT2(VOID)					// 21.04.98 cg, new, calculate timer2 value
{
	DWORD dwT2 = Chipset.t2;				// get value from chipset
	if (bStarted)							// timer2 running
	{
		LARGE_INTEGER lT2Act;
		QueryPerformanceCounter(&lT2Act);	// actual time
		// calculate ticks since reference point
		dwT2 -= (DWORD) 
				(((lT2Act.QuadPart - lT2Ref.QuadPart) * T2_FREQ) 
				/ lFreq.QuadPart);
	}
	return dwT2;
}

static VOID CheckT1(VOID)
{
	_ASSERT(bStarted);						// timer must run
	// 03.03.98 cg, new implementation of checking the timer1 control register
	if (Chipset.Shutdn)
	{
		// T1 Wake up, Interrupt, EXTRA
		if (Chipset.IORam[TIMER1_CTRL] & (WKE | INT | XTRA))
		{
			if (Chipset.IORam[TIMER1_CTRL] & (INT | XTRA))
			{
				// set service request and evaluate interrupt
				Chipset.IORam[TIMER1_CTRL] |= SRQ;
				Chipset.SoftInt = TRUE;
				bInterrupt=TRUE;
			}
			ResumeThread(hThread);			// wake up emulation thread
		}
	}
	else
	{
		// timer1 rollover and interrupt enable
		if (Chipset.t1 == 0x0f && (Chipset.IORam[TIMER1_CTRL] & INT))
		{
			Chipset.SoftInt = TRUE;			// evaluate interrupt
			bInterrupt = TRUE;
		}
	}
	// 03.03.98 cg, end of new implementation

	// if ((Chipset.t1&8) == 0)
	// {
	//	Chipset.IORam[0x2E] &= 0x7; // clear bit 3
	//	return;
	// }
	// if (Chipset.IORam[0x2E]&4)
	// { // T1 -> Wake Up
	//	if (Chipset.Shutdn)
	//	{
	//		ResumeThread(hThread);
	//		Chipset.IORam[0x2E] &= 0xB; // clear bit 2
	//		Chipset.IORam[0x2E] |= 0x8; // set bit 3
	//	}
	// }
	// if (Chipset.IORam[0x2E]&2)
	// { // T1 -> Interrupt
	//	Chipset.IORam[0x2E] |= 0x8; // set bit 3
	//	Chipset.SoftInt = TRUE;
	//	bInterrupt=TRUE;
	// }
	return;
}

static VOID CheckT2(VOID)
{
	_ASSERT(bStarted);						// timer must run
	if ((Chipset.t2&0x80000000) == 0)		// ???
	{
		Chipset.IORam[0x2F] &= 7;			// clear bit 3 (service request)
		return;
	}
	if (Chipset.IORam[0x2F]&4)
	{ // T2 -> Wake Up
		if (Chipset.Shutdn)
		{
			ResumeThread(hThread);
			Chipset.IORam[0x2F] &= 0xB;		// clear bit 2 (wake up)
			Chipset.IORam[0x2F] |= 0x8;		// set bit 3 (service request)
		}
	}
	if (Chipset.IORam[0x2F]&2)
	{ // T2 -> Interrupt
		Chipset.IORam[0x2F] |= 0x8;			// set bit 3 (service request)
		Chipset.SoftInt = TRUE;				// execute interrupt
		bInterrupt=TRUE;
	}
	return;
}

static VOID RescheduleT2(BOOL bRefPoint)	// 21.04.98 cg, add function parameter
{
	UINT uDelay;
	_ASSERT(uT2TimerId == 0);				// timer2 must stopped
	uDelay = Chipset.t2 >> 3;				// 21.04.98 cg, timer delay in ms
    uDelay = MAX(tc.wPeriodMin,uDelay);		// 21.04.98 cg, wait minimum delay of timer
	if (bOutRange = uDelay > tc.wPeriodMax)	// 21.04.98 cg, delay greater maximum delay
		uDelay = tc.wPeriodMax;				// 21.04.98 cg, wait maximum delay time
	// dwT2Init   = timeGetTime();			// 06.04.98 cg, not used any more
	// dwT2Ticks  = dwT2Init;				// 06.04.98 cg, not used any more
	if (bRefPoint)							// 21.04.98 cg, save reference time
		QueryPerformanceCounter(&lT2Ref);	// 06.04.98 cg, time of corresponding Chipset.t2 value
	// start timer2; schedule event, when Chipset.t2 will be zero (Chipset.t2 / 8 = time in ms)
	uT2TimerId = timeSetEvent(uDelay,0,(LPTIMECALLBACK)&TimeProc,2,TIME_ONESHOT);
	_ASSERT(uT2TimerId);					// 21.04.98 cg, new, test if timer2 started
	return;
}

static VOID AbortT2(VOID)
{
	_ASSERT(uT2TimerId);
	timeKillEvent(uT2TimerId);				// 09.10.97 cg, bugfix, first kill event
	uT2TimerId = 0;							// 09.10.97 cg, bugfix, then reset var
	return;
}

static void CALLBACK TimeProc(UINT uEventId, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
	_ASSERT(uEventId);						// 16.03.98 cg, new, illegal EventId
	if (!bStarted)							// timer stopped
		return;								// -> quit
	// called from timer1 event (default period 16 Hz)
	// if ((uT1TimerId!=0) && (uEventId == uT1TimerId))
	if (uEventId == uT1TimerId)				// 16.03.98 cg, removed useless part
	{
		EnterCriticalSection(&csT1Lock);	// 21.04.98 cg, bugfix, synchronize
		{
			Chipset.t1 = (Chipset.t1-1)&0xF;// decrement timer value
			CheckT1();						// test timer1 control bits
		}
		LeaveCriticalSection(&csT1Lock);	// 21.04.98 cg
		return;
	}
	// called from timer2 event, Chipset.t2 should be zero
	// if ((uT2TimerId!=0) && (uEventId == uT2TimerId))
	if (uEventId == uT2TimerId)				// 16.03.98 cg, removed useless part
	{
		uT2TimerId = 0;						// timer2 stopped, because it was a one shot timer
		EnterCriticalSection(&csT2Lock);	// 21.04.98 cg, bugfix, synchronize
		{
			if (!bOutRange)					// 21.04.98 cg, timer event elapsed
			{
				// 21.04.98 cg, changed, wait for timer overrun
				while(((Chipset.t2 = CalcT2()) & 0x80000000) == 0)
					Sleep(0);
				CheckT2();					// test timer2 control bits
			}
			RescheduleT2(!bOutRange);		// restart timer2
		}
		LeaveCriticalSection(&csT2Lock);	// 21.04.98 cg
		return;
	}
	return;
	UNREFERENCED_PARAMETER(uMsg);
	UNREFERENCED_PARAMETER(dwUser);			// 19.02.98 cg, added unreferenced parameter declarations
	UNREFERENCED_PARAMETER(dw1);
	UNREFERENCED_PARAMETER(dw2);
}


VOID StartTimers(VOID)
{
	if (bStarted)							// timer running
		return;								// -> quit
	if (Chipset.IORam[0x2F]&1)				// start timer2 ?
	{
		bStarted = TRUE;					// flag timer running
		InitializeCriticalSection(&csT1Lock);// 21.04.98 cg, bugfix, initialize critical section
		InitializeCriticalSection(&csT2Lock);// 21.04.98 cg, bugfix, initialize critical section
		SetAccesstime();					// 06.10.97 cg, set date and time
		if (bAccurateTimer)					// box "Accurate timer" checked
		{
			// 21.04.98 cg, optimized, set timer resolution to 1 ms, if failed don't use "Accurate timer"
			bAccurateTimer = (timeBeginPeriod(1) == TIMERR_NOERROR);
		}
		QueryPerformanceFrequency(&lFreq);	// 06.04.98 cg, init high resolution counter
		timeGetDevCaps(&tc,sizeof(tc));		// 21.04.98 cg, get timer resolution
		// set timer1 with given period
		uT1TimerId = timeSetEvent(uT1Period,0,(LPTIMECALLBACK)&TimeProc,1,TIME_PERIODIC);
		_ASSERT(uT1TimerId);				// 16.03.98 cg, new, test if timer1 started
		RescheduleT2(TRUE);					// start timer2
	}
	return;
}

VOID StopTimers(VOID)
{
	if (!bStarted)							// timer stopped
		return;								// -> quit
	// Chipset.t2 = ReadT2();				// 21.04.98 cg, removed, read timer2 value later
	if (uT1TimerId != 0)					// timer1 running
	{
		EnterCriticalSection(&csT1Lock);	// 21.04.98 cg, bugfix, synchronize
		{
			timeKillEvent(uT1TimerId);		// stop timer1
			uT1TimerId = 0;					// set flag timer1 stopped
		}
		LeaveCriticalSection(&csT1Lock);	// 21.04.98 cg
	}
	if (uT2TimerId != 0)					// timer2 running
	{
		EnterCriticalSection(&csT2Lock);	// 21.04.98 cg, bugfix, synchronize
		{
			Chipset.t2 = CalcT2();			// 21.04.98 cg, update chipset timer2 value
			AbortT2();						// stop timer2
		}
		LeaveCriticalSection(&csT2Lock);	// 21.04.98 cg
	}
	bStarted = FALSE;
	if (bAccurateTimer)						// "Accurate timer" running
	{
		timeEndPeriod(1);					// finish service
	}
	DeleteCriticalSection(&csT1Lock);		// 21.04.98 cg, bugfix, release critical section
	DeleteCriticalSection(&csT2Lock);		// 21.04.98 cg, bugfix, release critical section
	return;
}

DWORD ReadT2(VOID)
{
	// 21.04.98 cg, changed implementation
	DWORD dwT2;
	EnterCriticalSection(&csT2Lock);
	{
		dwT2 = CalcT2();					// calculate timer2 value
	}
	LeaveCriticalSection(&csT2Lock);
	return dwT2;
}

VOID SetT2(DWORD dwValue)
{
	// 21.04.98 cg, changed implementation
	EnterCriticalSection(&csT2Lock);
	{
		if (uT2TimerId != 0)				// timer2 running
			AbortT2();						// stop timer2
		Chipset.t2 = dwValue;				// set new value
		CheckT2();							// test timer2 control bits
		if (bStarted)						// timer running
			RescheduleT2(TRUE);				// restart timer2
	}
	LeaveCriticalSection(&csT2Lock);
	return;
}

BYTE ReadT1(VOID)
{
	BYTE nT1;
	EnterCriticalSection(&csT1Lock);		// 21.04.98 cg, bugfix, synchronize
	{
		nT1 = Chipset.t1;					// read timer1 value
	}
	LeaveCriticalSection(&csT1Lock);		// 21.04.98 cg
	return nT1;
}

VOID SetT1(BYTE byValue)
{
	EnterCriticalSection(&csT1Lock);		// 21.04.98 cg, bugfix, synchronize
	{
		Chipset.t1 = byValue&0xF;			// set new timer1 value
		CheckT1();							// test timer1 control bits
	}
	LeaveCriticalSection(&csT1Lock);		// 21.04.98 cg
	return;
}
